// Test code for aes_core.c replacement for EZchip TILE-Gx.
// Written by Nils Liaaen Corneliusen 2015.
// Please refer to the article at https://www.ignorantus.com/pages/tilegx_aes_openssl/
// for more information.

// tile-cc  -Wall -o test_aes -O3 -std=gnu99 -static --save-temps test.c aes_core.c aes_core_tilegx.c
// tile-cc -DGX_STORE_ALIGNED -Wall -o test_aes -O3 -std=gnu99 -static --save-temps test.c aes_core.c aes_core_tilegx.c
// tile-monitor --simulator --image 2x2 --upload test_aes /tmp/aes_gx --run -+- /tmp/aes_gx -+- --quit --functional

#include <stdio.h>
#include <stdbool.h>
#include <stdint.h>
#include <arch/cycle.h>

#include "aes.h"

static uint32_t rand32( void )
{
    return ((rand()&0xffff)<<16) | rand();
}

static uint64_t rand64( void )
{
    return (((uint64_t)rand32())<<32) | rand();
}

static bool memc( uint8_t *buf, uint8_t c, int len )
{
    for( int i = 0; i < len; i++ ) {
        if( buf[i] != c ) return true;
    }
    return false;
}

static __attribute__ ((noinline)) uint64_t clockish( void )
{
    return get_cycle_count();
//    struct timeval tv;
//    gettimeofday( &tv, NULL );
//    return (((uint64_t)tv.tv_sec)*1000000)+((uint64_t)tv.tv_usec);
}

static void test_enc( int rounds, AES_KEY *aesKey )
{
    uint64_t src[16];
    uint64_t dsto[16];
    uint64_t dstn[16];

    printf( "enc%d ", rounds );

    for( int i = 0; i < 16; i++ ) {
        src[i] = rand64();
    }

    bool error = false;

    for( int i = 0; i < 8; i++ ) {
        uint8_t *src8 = ((uint8_t *)src)+i;

        printf( "%d", i );

        memset( dsto, 0xff, sizeof(dsto) );
        memset( dstn, 0xff, sizeof(dstn) );

        aesKey->rounds = rounds;

        AES_encrypt(        src8, (uint8_t *)dsto, aesKey );
        AES_encrypt_tilegx( src8, (uint8_t *)dstn, aesKey );

        if( memcmp( (uint8_t *)dsto,     dstn, 16 ) ) { printf( " fail " );    error = true; }
        if( memc(   (uint8_t *)(dstn+2), 0xff, 16 ) ) { printf( " overrun " ); error = true; }

    }

    if( !error) printf( " ok\n" );
}

static void test_dec( int rounds, AES_KEY *aesKey )
{
    uint64_t src[16];
    uint64_t dsto[16];
    uint64_t dstn[16];

    printf( "dec%d ", rounds );

    for( int i = 0; i < 16; i++ ) {
        src[i] = rand64();
    }

    bool error = false;

    for( int i = 0; i < 8; i++ ) {
        uint8_t *src8 = ((uint8_t *)src)+i;

        printf( "%d", i );

        memset( dsto, 0xff, sizeof(dsto) );
        memset( dstn, 0xff, sizeof(dstn) );

        aesKey->rounds = rounds;

        AES_decrypt(        src8, (uint8_t *)dsto, aesKey );
        AES_decrypt_tilegx( src8, (uint8_t *)dstn, aesKey );

        if( memcmp( (uint8_t *)dsto,     dstn, 16 ) ) { printf( " fail " );    error = true; }
        if( memc(   (uint8_t *)(dstn+2), 0xff, 16 ) ) { printf( " overrun " ); error = true; }

    }

    if( !error) printf( " ok\n" );
}

#define RUNS 20000

int main( int argc, char *argv[] )
{
    AES_KEY aesKey;
    uint64_t src[16];
    uint64_t dsto[16];
    uint64_t dstn[16];

    srand( 42 );

    for( int i = 0; i < 4*14; i++ ) {
        aesKey.rd_key[i] = rand32();
    }

    for( int i = 0; i < 16; i++ ) {
        src[i] = rand64();
    }

    test_enc( 10, &aesKey );
    test_enc( 12, &aesKey );
    test_enc( 14, &aesKey );

    test_dec( 10, &aesKey );
    test_dec( 12, &aesKey );
    test_dec( 14, &aesKey );

    aesKey.rounds = 10;

    uint64_t t0,t1,t2,t3;
    int tn,to;

    printf( "AES_Encrypt()\n" );

    t0 = clockish();
    for( int i = 0; i < RUNS; i++ ) {
        AES_encrypt( (uint8_t *)src,  (uint8_t *)dsto, &aesKey );
        AES_encrypt( (uint8_t *)dsto, (uint8_t *)dstn, &aesKey );
    }
    t1 = clockish();

    to = t1-t0;

    t2 = clockish();
    for( int i = 0; i < RUNS; i++ ) {
        AES_encrypt_tilegx( (uint8_t *)src,  (uint8_t *)dsto, &aesKey );
        AES_encrypt_tilegx( (uint8_t *)dsto, (uint8_t *)dstn, &aesKey );
    }
    t3 = clockish();

    tn = t3-t2;

    printf( "%d %d %f%%\n", to, tn, (float)(to-tn)*100.0f/(float)to );

    printf( "AES_Decrypt()\n" );

    t0 = clockish();
    for( int i = 0; i < RUNS; i++ ) {
        AES_decrypt( (uint8_t *)src,  (uint8_t *)dsto, &aesKey );
        AES_decrypt( (uint8_t *)dsto, (uint8_t *)dstn, &aesKey );
    }
    t1 = clockish();

    to = t1-t0;

    t2 = clockish();
    for( int i = 0; i < RUNS; i++ ) {
        AES_decrypt_tilegx( (uint8_t *)src,  (uint8_t *)dsto, &aesKey );
        AES_decrypt_tilegx( (uint8_t *)dsto, (uint8_t *)dstn, &aesKey );
    }
    t3 = clockish();

    tn = t3-t2;

    printf( "%d %d %f%%\n", to, tn, (float)(to-tn)*100.0f/(float)to );

    printf( "done\n" );
    return 0;
}
